/*
 *
 * Copyright (C) 2008 Tomasz Ostrowski ( http://ostry.w.szu.pl )
 *
 * This file is part of Free GG.
 *
 * Free GG is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Free GG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Free GG; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "PluginManager.h"
#include "GaduThread.h"
#include "Main.h"
#include "StatusDesc.h"

extern TGaduThread *gg_thread;

PluginManager *PlugManager = NULL;
HANDLE      MessageMutex;
HANDLE		MessageThread;
PLUGIN_MSG  MessageQueue;
bool MessageThreadIsRunning = false;
DWORD WINAPI MessageThreadProc(LPVOID data)
{
	while (MessageThreadIsRunning)
	{
		PlugManager->HandleNextMessage();
		Sleep(10);
	}
	return 0;
}

PluginManager::PluginManager(char* dirname, char* inidirname)
{
    PluginCount = 0;
	MessageMutex = CreateMutex(NULL,FALSE,NULL);

	LoadPluginsFromDir(dirname);
	char *broadcast = new char[strlen("CONFIG INIPATH ") + strlen(inidirname) + 1];
	strcpy(broadcast, "CONFIG INIPATH ");
	strcat(broadcast, inidirname);
	PostMsg(broadcast);
	delete [] broadcast;
}

PluginManager::~PluginManager()
{
	MessageThreadIsRunning = false;
	DestroyPlugins();
	TerminateThread(MessageThread, 0);
}


int PluginManager::LoadPluginsFromDir(char* dirname)
{
	WIN32_FIND_DATA file;
	char *srchpath = new char[strlen(dirname) + strlen("*.dll") + 1];
	strcpy(srchpath, dirname);
	strcat(srchpath, "*.dll");
	HANDLE hFind = FindFirstFile(srchpath,&file);
	int hasfiles = (hFind != INVALID_HANDLE_VALUE);

	while (hasfiles && PluginCount <= MAX_PLUGIN_COUNT)
	{
		char *filename = new char[strlen(dirname) + strlen(file.cFileName) + 1];
		strcpy(filename, dirname);
		strcat(filename, file.cFileName);
		LoadPlugin(filename);
		delete [] filename;
		hasfiles = FindNextFile(hFind, &file);
	}
	FindClose(hFind);
	delete [] srchpath;
	return 0;
}

int PluginManager::Run(void)
{
	MessageThreadIsRunning = true;
	MessageThread = CreateThread(NULL,0,MessageThreadProc,NULL,0,&dwtid);
	return 0;
}

int PluginManager::LoadPlugin(char* filename)
{
    PLUGIN plugin;
    plugin.hInstance = LoadLibrary(filename);
    if (!plugin.hInstance) return -1;
    
    plugin.Load =   (PROC_LOAD)     GetProcAddress(plugin.hInstance, "Load");
    plugin.Unload = (PROC_UNLOAD)   GetProcAddress(plugin.hInstance, "Unload");
    plugin.Name =   (PROC_NAME)     GetProcAddress(plugin.hInstance, "Name");

    plugin.HandleMessage =  (PROC_HANDLEMSG) GetProcAddress(plugin.hInstance, "HandleMessage");
    
    if (plugin.Load == NULL ||
        plugin.Unload == NULL ||
        plugin.Name == NULL ||
        plugin.HandleMessage == NULL) 
    {
        return -2;
    }
    
    PluginList[PluginCount] = plugin;
    PluginCount++;
    plugin.Load(QueueMessage);
    return 0;
}

int PluginManager::DestroyPlugins()
{
    while (PluginCount)
    {
        PluginCount--;
        PluginList[PluginCount].Unload();
        FreeLibrary(PluginList[PluginCount].hInstance);
	}
	return 0;
}


int QueueMessage(char* plugin, char* message)
{
    if (WaitForSingleObject(MessageMutex, 5000L) != WAIT_OBJECT_0) return 1;
    
    PLUGIN_MSG *pointer = &MessageQueue;
    while (pointer->NextMessage) pointer = pointer->NextMessage;
    
    pointer->NextMessage = (PLUGIN_MSG*) malloc(sizeof(PLUGIN_MSG));
    pointer = pointer->NextMessage;
    
    int len_name = strlen(plugin)+1;
    int len_mesg = strlen(message)+1;
    
    pointer->PluginName = (char*) malloc(len_name);
    pointer->Message    = (char*) malloc(len_mesg);
    
    memcpy(pointer->PluginName, plugin,  len_name);
    memcpy(pointer->Message   , message, len_mesg);
    
    pointer->NextMessage = 0;
    ReleaseMutex(MessageMutex);
   
    return 0;
}

/** Obsluz wiadomosc od pluginu
	<zarys funkcji>
*/
int PluginManager::HandleNextMessage()
{
    int nr = 0;
    
    if (WaitForSingleObject(MessageMutex, 5000L) != WAIT_OBJECT_0) return -1;
    PLUGIN_MSG *pointer = MessageQueue.NextMessage;
    if (pointer) MessageQueue.NextMessage = pointer->NextMessage;
    ReleaseMutex(MessageMutex);

    if (!pointer) return 0;
	char *pMsg = pointer->Message;

	//Handle the message internally
	if (!stricmp(pointer->PluginName, "freegg"))
	{
		if (StartsWith(pMsg, "set "))
        {
			if (StartsWith(pMsg, "status "))
			{
				if (StartsWith(pMsg, "mode "))
				{

				}
				else if (StartsWith(pMsg, "desc "))
				{
					MainForm->ChangeStatus (-1, pMsg);
				}
			}
		}
		else if (StartsWith(pMsg, "get "))
		{
			if (StartsWith(pMsg, "status "))
			{
				if (StartsWith(pMsg, "mode"))
				{

				}
				else if (StartsWith(pMsg, "desc"))
				{

				}
			}
		}
	}
    else
    {
		for (int i=0; i<PluginCount; i++)
		{
			if (!strcmp(PluginList[i].Name(), pointer->PluginName))
			{
				PluginList[i].HandleMessage(pointer->Message);
				nr++;
			}
		}
    }
    free(pointer->PluginName);
    free(pointer->Message);
    free(pointer);
            
    return nr;
}

/** Send broadcast message to all plugins
*/
void PluginManager::PostMsg(char* msg)
{
	for (int i=0; i<PluginCount; i++)
	{
		PluginList[i].HandleMessage(msg);
	}
}

/** Show plugin with iID position in PluginList.
	Plugin may show its config window or anything else.
*/
void PluginManager::ShowPlugin(int iID)
{
	PluginList[iID].HandleMessage("SHOW MAIN");
}

inline bool PluginManager::StartsWith(char* &a, char *b)
{
	if(strlen(a) < strlen(b))
	{
		return false;
	}
	else
	{
		bool result = !strnicmp(a,b,strlen(b));
		if(result)
        	a += strlen(b);
		return result;
	}
}
